==> v10.5.0.tar.gz: brewman/brewman/brewman/routers/reports/stock_movement.py <==
from datetime import date, datetime
from decimal import Decimal
from typing import List

import brewman.schemas.stock_movement as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session, contains_eager
from sqlalchemy.sql.expression import func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.batch import Batch
from ...models.cost_centre import CostCentre
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.product import Product
from ...models.product_group import ProductGroup
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.StockMovement)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["stock-movement"]),
) -> schemas.StockMovement:
    return schemas.StockMovement(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        body=[],
    )


@router.get("/{start}/{finish}", response_model=schemas.StockMovement)
def report_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["stock-movement"]),
) -> schemas.StockMovement:
    with SessionFuture() as db:
        body = build_stock_movement(datetime.strptime(start, "%d-%b-%Y"), datetime.strptime(finish, "%d-%b-%Y"), db)
        set_period(start, finish, request.session)
        return schemas.StockMovement(startDate=start, finishDate=finish, body=body)


def build_stock_movement(start_date: date, finish_date: date, db: Session) -> List[schemas.StockMovementItem]:
    dict_ = {}
    quantity_sum = func.sum(Journal.debit * Inventory.quantity).label("quantity")
    openings = db.execute(
        select(StockKeepingUnit, quantity_sum)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(StockKeepingUnit.batches)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .options(contains_eager(StockKeepingUnit.product).contains_eager(Product.product_group))
        .where(Voucher.date < start_date, Journal.cost_centre_id == CostCentre.cost_centre_purchase())
        .group_by(StockKeepingUnit, Product, ProductGroup)
    ).all()
    for sku, quantity in openings:
        dict_[sku.id] = schemas.StockMovementItem(
            id=sku.id,
            name=f"{sku.product.name} ({sku.units})",
            group=sku.product.product_group.name,
            opening=Decimal(round(quantity, 2)),
            purchase=0,
            issue=0,
            closing=0,
            url=["/", "product-ledger", str(sku.product.id)],
        )
    purchases = db.execute(
        select(StockKeepingUnit, quantity_sum)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(StockKeepingUnit.batches)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .options(contains_eager(StockKeepingUnit.product).contains_eager(Product.product_group))
        .where(
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
        .group_by(StockKeepingUnit, Product, ProductGroup)
    ).all()
    for sku, quantity in purchases:
        if sku.id in dict_:
            dict_[sku.id].purchase = Decimal(round(quantity, 2))
        else:
            dict_[sku.id] = schemas.StockMovementItem(
                id=sku.id,
                name=f"{sku.product.name} ({sku.units})",
                group=sku.product.product_group.name,
                opening=0,
                purchase=Decimal(round(quantity, 2)),
                issue=0,
                closing=0,
                url=["/", "product-ledger", str(sku.product.id)],
            )
    issues = db.execute(
        select(StockKeepingUnit, quantity_sum)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(StockKeepingUnit.batches)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .options(contains_eager(StockKeepingUnit.product).contains_eager(Product.product_group))
        .where(
            Voucher.date >= start_date,
            Voucher.date <= finish_date,
            Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK]),
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
        .group_by(StockKeepingUnit, Product, ProductGroup)
    ).all()
    for sku, quantity in issues:
        if sku.id in dict_:
            dict_[sku.id].issue = Decimal(round(quantity * -1, 2))
        else:
            dict_[sku.id] = schemas.StockMovementItem(
                id=sku.id,
                name=f"{sku.product.name} ({sku.units})",
                group=sku.product.product_group.name,
                opening=0,
                purchase=0,
                issue=Decimal(round(quantity * -1, 2)),
                closing=0,
                url=["/", "product-ledger", str(sku.product.id)],
            )

    list_ = [value for key, value in dict_.items()]
    list_ = sorted(list_, key=lambda x: x.name.lower())
    list_ = sorted(list_, key=lambda x: x.group.lower())
    for i in range(len(list_), 0, -1):
        item = list_[i - 1]
        if item.opening == 0 and item.purchase == 0 and item.issue == 0:
            list_.remove(item)
        else:
            item.closing = round(item.opening + item.purchase - item.issue, 2)
    return list_

==> v11.1.2.tar.gz: brewman/brewman/brewman/routers/reports/stock_movement.py <==
from datetime import date, datetime
from decimal import Decimal

import brewman.schemas.stock_movement as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session, contains_eager
from sqlalchemy.sql.expression import func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.batch import Batch
from ...models.cost_centre import CostCentre
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.product import Product
from ...models.product_group import ProductGroup
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.StockMovement)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["stock-movement"]),
) -> schemas.StockMovement:
    return schemas.StockMovement(
        start_date=get_start_date(request.session),
        finish_date=get_finish_date(request.session),
        body=[],
    )


@router.get("/{start}/{finish}", response_model=schemas.StockMovement)
def report_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["stock-movement"]),
) -> schemas.StockMovement:
    with SessionFuture() as db:
        body = build_stock_movement(datetime.strptime(start, "%d-%b-%Y"), datetime.strptime(finish, "%d-%b-%Y"), db)
        set_period(start, finish, request.session)
        return schemas.StockMovement(start_date=start, finish_date=finish, body=body)


def build_stock_movement(start_date: date, finish_date: date, db: Session) -> list[schemas.StockMovementItem]:
    dict_ = {}
    quantity_sum = func.sum(Journal.debit * Inventory.quantity).label("quantity")
    openings = db.execute(
        select(StockKeepingUnit, quantity_sum)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(StockKeepingUnit.batches)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .options(contains_eager(StockKeepingUnit.product).contains_eager(Product.product_group))
        .where(Voucher.date_ < start_date, Journal.cost_centre_id == CostCentre.cost_centre_purchase())
        .group_by(StockKeepingUnit, Product, ProductGroup)  # type: ignore
    ).all()
    for sku, quantity in openings:
        dict_[sku.id] = schemas.StockMovementItem(
            id_=sku.id,
            name=f"{sku.product.name} ({sku.units})",
            group=sku.product.product_group.name,
            opening=Decimal(round(quantity, 2)),
            purchase=Decimal(0),
            issue=Decimal(0),
            closing=Decimal(0),
            url=["/", "product-ledger", str(sku.product.id)],
        )
    purchases = db.execute(
        select(StockKeepingUnit, quantity_sum)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(StockKeepingUnit.batches)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .options(contains_eager(StockKeepingUnit.product).contains_eager(Product.product_group))
        .where(
            Voucher.date_ >= start_date,
            Voucher.date_ <= finish_date,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
        .group_by(StockKeepingUnit, Product, ProductGroup)  # type: ignore
    ).all()
    for sku, quantity in purchases:
        if sku.id in dict_:
            dict_[sku.id].purchase = Decimal(round(quantity, 2))
        else:
            dict_[sku.id] = schemas.StockMovementItem(
                id_=sku.id,
                name=f"{sku.product.name} ({sku.units})",
                group=sku.product.product_group.name,
                opening=Decimal(0),
                purchase=Decimal(round(quantity, 2)),
                issue=Decimal(0),
                closing=Decimal(0),
                url=["/", "product-ledger", str(sku.product.id)],
            )
    issues = db.execute(
        select(StockKeepingUnit, quantity_sum)
        .join(StockKeepingUnit.product)
        .join(Product.product_group)
        .join(StockKeepingUnit.batches)
        .join(Batch.inventories)
        .join(Inventory.voucher)
        .join(Voucher.journals)
        .options(contains_eager(StockKeepingUnit.product).contains_eager(Product.product_group))
        .where(
            Voucher.date_ >= start_date,
            Voucher.date_ <= finish_date,
            Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK]),
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
        .group_by(StockKeepingUnit, Product, ProductGroup)  # type: ignore
    ).all()
    for sku, quantity in issues:
        if sku.id in dict_:
            dict_[sku.id].issue = Decimal(round(quantity * -1, 2))
        else:
            dict_[sku.id] = schemas.StockMovementItem(
                id_=sku.id,
                name=f"{sku.product.name} ({sku.units})",
                group=sku.product.product_group.name,
                opening=Decimal(0),
                purchase=Decimal(0),
                issue=Decimal(round(quantity * -1, 2)),
                closing=Decimal(0),
                url=["/", "product-ledger", str(sku.product.id)],
            )

    list_ = [value for key, value in dict_.items()]
    list_ = sorted(list_, key=lambda x: x.name.lower())
    list_ = sorted(list_, key=lambda x: x.group.lower())
    for i in range(len(list_), 0, -1):
        item = list_[i - 1]
        if item.opening == 0 and item.purchase == 0 and item.issue == 0:
            list_.remove(item)
        else:
            item.closing = round(item.opening + item.purchase - item.issue, 2)
    return list_
